from django.conf import settings
from typing import Any
from .utils.extras import get_if_present


_config = {
    # General
    "APP_NAME": get_if_present(settings.AXOR_AUTH, 'APP_NAME', 'Django Axor Auth'),
    "URI_PREFIX": get_if_present(settings.AXOR_AUTH, 'URI_PREFIX', '/api'),
    "FRONTEND_URL": get_if_present(settings.AXOR_AUTH, 'FRONTEND_URL', None),

    # Cookies
    "AUTH_COOKIE_NAME": get_if_present(settings.AXOR_AUTH, 'AUTH_COOKIE_NAME', 'axor_auth'),
    # 1 week
    "AUTH_COOKIE_AGE": get_if_present(settings.AXOR_AUTH, 'AUTH_COOKIE_AGE', 60 * 60 * 24 * 7),
    "AUTH_COOKIE_SECURE": get_if_present(settings.AXOR_AUTH, 'AUTH_COOKIE_SECURE', False),
    "AUTH_COOKIE_SAMESITE": get_if_present(settings.AXOR_AUTH, 'AUTH_COOKIE_SAMESITE', 'SameSite'),
    "AUTH_COOKIE_DOMAIN": get_if_present(settings.AXOR_AUTH, 'AUTH_COOKIE_DOMAIN', None),

    # Forgot Password
    # 30 mins
    "FORGET_PASSWORD_LINK_TIMEOUT": get_if_present(settings.AXOR_AUTH, 'FORGET_PASSWORD_LINK_TIMEOUT', 30),
    # 24 hrs
    "FORGET_PASSWORD_LOCKOUT_TIME": get_if_present(settings.AXOR_AUTH, 'FORGET_PASSWORD_LOCKOUT_TIME', 24),

    # TOTP
    "TOTP_NUM_OF_BACKUP_CODES": get_if_present(settings.AXOR_AUTH, 'TOTP_NUM_OF_BACKUP_CODES', 8),
    "TOTP_BACKUP_CODE_LENGTH": get_if_present(settings.AXOR_AUTH, 'TOTP_BACKUP_CODE_LENGTH', 8),

    # SMTP
    "SMTP_USE_TLS": get_if_present(settings.AXOR_AUTH, 'SMTP_USE_TLS', True),
    "SMTP_USE_SSL": get_if_present(settings.AXOR_AUTH, 'SMTP_USE_SSL', False),
    "SMTP_HOST": get_if_present(settings.AXOR_AUTH, 'SMTP_HOST', None),
    "SMTP_PORT": get_if_present(settings.AXOR_AUTH, 'SMTP_PORT', None),
    "SMTP_USER": get_if_present(settings.AXOR_AUTH, 'SMTP_USER', None),
    "SMTP_PASSWORD": get_if_present(settings.AXOR_AUTH, 'SMTP_PASSWORD', None),
    "SMTP_DEFAULT_SEND_FROM": get_if_present(settings.AXOR_AUTH, 'SMTP_DEFAULT_SEND_FROM', None),
}


class Config:
    APP_NAME: str
    URI_PREFIX: str
    FRONTEND_URL: str
    AUTH_COOKIE_NAME: str
    AUTH_COOKIE_AGE: int
    AUTH_COOKIE_SECURE: bool
    AUTH_COOKIE_SAMESITE: str
    AUTH_COOKIE_DOMAIN: str
    FORGET_PASSWORD_LINK_TIMEOUT: int
    FORGET_PASSWORD_LOCKOUT_TIME: int
    TOTP_NUM_OF_BACKUP_CODES: int
    TOTP_BACKUP_CODE_LENGTH: int
    SMTP_USE_TLS: bool
    SMTP_USE_SSL: bool
    SMTP_HOST: str
    SMTP_PORT: str
    SMTP_USER: str
    SMTP_PASSWORD: str
    SMTP_DEFAULT_SEND_FROM: str

    def __init__(self, data: dict[str, Any]):
        for key, value in data.items():
            setattr(self, key, value)


# Create an instance of Config with the dictionary data
config = Config(_config)



from .....utils.emailing.base_template import base_template


def reset_password_success_template(first_name, email_subject='Password was changed'):
    return base_template(
        subject=email_subject,
        headline="Password was changed",
        message=[
            'Hi ' + first_name + ',',
            'Your password was successfully changed. If you did not make this change, please contact our support immediately.',
            '',
            'https://support.plutohealth.com/'
            '',
            'Thank you for using our services.',
        ]
    )



from django.utils.encoding import force_str
from rest_framework.decorators import api_view
from rest_framework.response import Response
from ...utils.error_handling.error_message import ErrorMessage
from ..serializers import PasswordSerializer
from ..users_utils.emailing.api import send_forgot_password_email, send_password_changed_email
from ..models import UserPasswordChange
from .models import ForgotPassword
from .serializers import HealthyForgotPasswordSerializer
from .utils import getClientIP
from django_axor_auth.configurator import config


@api_view(['POST'])
def forgot_password(request):
    # Check if all required fields are provided
    if 'email' not in request.data:
        err = ErrorMessage(
            title='Email is required.',
            status=400,
            detail='Email is required.',
            instance=request.get_full_path(),
        )
        return err.to_response()
    key, fp = ForgotPassword.objects.create_forgot_password(
        request,
        force_str(request.data['email'])
    )
    if key and fp:
        # Send email with key
        url = f'{config.FRONTEND_URL}/forgot-password/complete?token={key}'
        send_forgot_password_email(
            email=force_str(fp.user.email),
            reset_url=url,
            first_name=fp.user.first_name,
            subject='Reset your password',
            ip=getClientIP(request),
        )
    # Send empty success response
    return Response(status=200)


@api_view(['POST'])
def check_health(request):
    key = force_str(request.data['key'])
    serializer = HealthyForgotPasswordSerializer(data={'key': key})
    if not serializer.is_valid():
        err = ErrorMessage(
            title='Invalid Request',
            status=400,
            detail=serializer.errors,
            instance=request.get_full_path(),
        )
        return err.to_response()
    # Send empty success response
    return Response(status=200)


@api_view(['POST'])
def reset_password(request):
    # Check if password is provided
    if 'password' not in request.data:
        err = ErrorMessage(
            title='Password Required',
            status=400,
            detail='Password is required.',
            instance=request.get_full_path(),
            code="InsufficientData"
        )
        return err.to_response()
    key = force_str(request.data['key'])
    # Check if key is valid
    serializer = HealthyForgotPasswordSerializer(data={'key': key})
    if not serializer.is_valid():
        err = ErrorMessage(
            title='Invalid Request',
            status=400,
            detail=serializer.errors,
            instance=request.get_full_path(),
            code="HealthCheckFailed"
        )
        return err.to_response()
    fp = serializer.validated_data
    # Validate password
    serializer = PasswordSerializer(data=request.data)
    if serializer.is_valid():
        # Set new password
        fp.user.set_password(force_str(request.data['password']))
        fp.user.save()
        # Set fp as used
        fp.set_used()
        # Send notification email to user
        send_password_changed_email(
            email=fp.user.email,
            first_name=fp.user.first_name,
            subject='Password Changed'
        )
        # Update Password Change table for user
        UserPasswordChange.objects.save_due_forgot_password_form(fp.user)
        # Send empty success response
        return Response(status=204)
    else:
        err = ErrorMessage(
            title='Encountered Error',
            status=400,
            detail=serializer.errors,
            instance=request.get_full_path(),
        )
        return err.to_response()



from .models import AppToken
from rest_framework import serializers


class AppTokenSerializer(serializers.ModelSerializer):
    class Meta:
        model = AppToken
        fields = ['id', 'token', 'user', 'is_valid', 'created_at', 'ip', 'ua']
        extra_kwargs = {
            'token': {'required': True, 'write_only': True},
            'user': {'required': True},
        }


class UserAppTokenSerializer(serializers.ModelSerializer):
    class Meta:
        model = AppToken
        fields = ['id', 'is_valid', 'created_at', 'ip', 'ua']



from django.core.mail import send_mail
from .templates.new_user_account import new_user_account_template
from .templates.forgot_password import forgot_password_template
from .templates.reset_password_success import reset_password_success_template
from django_axor_auth.utils.emailing.helper import getEmailConnection
from django_axor_auth.configurator import config


def send_welcome_email(first_name: str, verification_url: str, email: str, subject='Welcome to Pluto Health'):
    """Send a welcome email to a new user

    Args:
        `first_name` (str): Users first name
        `verification_key` (str): Verification token
        `email` (str): Recipient email address.
        `subject` (str, optional): Email subject. Defaults to 'Welcome to Pluto Health'.

    Returns:
        `int`: Returns 0 if fails.
    """
    template = new_user_account_template(first_name, verification_url, subject)
    try:
        with getEmailConnection() as connection:
            message = f"Hello {first_name},\n\nPlease verify your email by clicking the link below.\n\n{verification_url}"
            return send_mail(subject=subject,
                             message=message,
                             html_message=template,
                             from_email=config.SMTP_DEFAULT_SEND_FROM,
                             recipient_list=[email,],
                             connection=connection)
    except Exception as e:
        return 0


def send_forgot_password_email(first_name: str, reset_url: str, email: str, ip: str, subject='Reset your password'):
    """Send a welcome email to a new user

    Args:
        `first_name` (str): Users first name
        `verification_key` (str): Verification token
        `email` (str): Recipient email address.
        `ip` (str): IP address of the user requesting password reset.
        `subject` (str, optional): Email subject. Defaults to 'Welcome to Pluto Health'.

    Returns:
        `int`: Returns 0 if fails.
    """
    template = forgot_password_template(first_name, reset_url, ip, subject)
    try:
        with getEmailConnection() as connection:
            message = f"Hello {first_name},\n\nPlease reset your password by clicking the link below.\n\n{reset_url}"
            return send_mail(subject=subject,
                             message=message,
                             html_message=template,
                             from_email=config.SMTP_DEFAULT_SEND_FROM,
                             recipient_list=[email,],
                             connection=connection)
    except Exception as e:
        return 0


def send_password_changed_email(first_name: str, email: str, subject='Password was changed'):
    """Send a welcome email to a new user

    Args:
        `first_name` (str): Users first name
        `subject` (str, optional): Email subject. Defaults to 'Welcome to Pluto Health'.

    Returns:
        `int`: Returns 0 if fails.
    """
    template = reset_password_success_template(first_name, subject)
    try:
        with getEmailConnection() as connection:
            message = f"Hello {first_name},\n\nYour password was successfully changed. If you did not make this change, please contact our support immediately."
            return send_mail(subject=subject,
                             message=message,
                             html_message=template,
                             from_email=config.SMTP_DEFAULT_SEND_FROM,
                             recipient_list=[email,],
                             connection=connection)
    except Exception as e:
        return 0



from django.http import JsonResponse

# https://datatracker.ietf.org/doc/html/rfc7807


class ErrorMessage():
    """Error message class that follows RFC7807 standard. It ignores `type` field and uses `code` instead.

    Attributes:
        title (str): A short summary of the problem type.
        detail (str): Explanation specific to this occurrence of the problem.
        status (int): HTTP code, e.g., 400, 404, 500. Use `status` from `rest_framework`.
        instance (str): URL that faced the problem. Get this from request.
        code (str): Custom code to identify the error. Optional.

    Usage:
        ```
        from utils import error_handling
        from utils.error_handling.error_message import ErrorMessage
        error = ErrorMessage(
            title='Incorrect credentials',
            detail='Username or password is incorrect',
            status=400,
            instance=request.build_absolute_uri(),
            code='AUTH400'
        )
        return error.to_response()
        ```
    """

    def __init__(self, title, detail, status, instance, code=None):
        # err_type is a URL that defines the problem type.
        # title is a short summary of the problem type.
        self.title = title
        # detail is explanation specific to this occurrence of the problem.
        self.detail = detail
        # status is the HTTP code
        self.status = status
        # instance is the URL that faced the problem
        self.instance = instance
        self.code = code

    def to_response(self):
        response = JsonResponse({
            'title': self.title,
            'detail': self.detail,
            'instance': self.instance,
            'code': self.code
        }, status=self.status)
        # Content type: problem+json
        response['Content-Type'] = 'application/problem+json'
        return response

    def serialize(self):
        return {
            'title': self.title,
            'detail': self.detail,
            'instance': self.instance,
            'code': self.code
        }



from django.utils.encoding import force_str
from rest_framework import serializers
from .models import ForgotPassword
from .utils import hash_this


class HealthyForgotPasswordSerializer(serializers.Serializer):
    key = serializers.CharField()

    def validate(self, data):
        err = 'Link or instance is no longer valid. Please request a new one.'
        # Check if key is provided
        if 'key' not in data:
            raise serializers.ValidationError(
                'Key/token is not present.')
        try:
            fp = ForgotPassword.objects.select_related('user').get(
                key=hash_this(force_str(data['key'])))
            # Check if key is correct
            if not fp.check_key(force_str(data['key'])):
                raise serializers.ValidationError(err)
            # Check if fp is valid
            if not fp.check_valid():
                raise serializers.ValidationError(err)
        except ForgotPassword.DoesNotExist:
            raise serializers.ValidationError(err)
        return fp



from django.contrib import admin
from .models import Session

# Register your models here.


class SessionAdmin(admin.ModelAdmin):
    list_display = ('id', 'user', 'is_valid', 'ip', 'created_at')
    search_fields = ['user__email', 'user__first_name',
                     'user__last_name', 'id', 'ip']
    search_help_text = 'Search with user\'s first and last name, email, session Id, or IP address.'
    list_per_page = 100
    ordering = ('-created_at',)
    save_on_top = False
    save_as = False
    readonly_fields = ['id', 'user', 'key', 'ip', 'ua',
                       'created_at', 'updated_at', 'expire_at']

    def has_delete_permission(self, request, obj=None):
        return False

    def has_add_permission(self, request):
        return False


admin.site.register(Session, SessionAdmin)



import uuid
from django.db import models
from ..models import User
from django.utils.timezone import now
from django.db import models
from .managers import AppTokenManager


class AppToken(models.Model):
    id = models.UUIDField(primary_key=True, default=uuid.uuid4)
    user = models.ForeignKey(User, on_delete=models.CASCADE)
    token = models.CharField(max_length=64)
    is_valid = models.BooleanField(default=True)
    ip = models.GenericIPAddressField()
    ua = models.TextField()
    created_at = models.DateTimeField(default=now)

    class Meta:
        db_table = 'axor_app_tokens'
        ordering = ['-created_at']

    objects = AppTokenManager()

    def __str__(self):
        return f"{self.pk}"



import pyotp
from datetime import timedelta
from django.db import models
from django.utils.timezone import now
from .utils import generate_backup_codes
from ...security.encryption import encrypt, decrypt
from django.db.models import Q


class TotpManager(models.Manager):
    def __init__(self):
        super().__init__()
        # Timeout when incorrect backup codes are
        # entered more than 3 times
        self.BC_ATTEMPTS = 3
        self.BC_TIMEOUT = timedelta(minutes=30)

    def create_totp(self, user):
        # Get the user's totp objects. It should ideally be only one.
        # If status is enabled or backup_used, then user already has totp enabled
        # We do not want to create a new totp object
        totp_row = self.filter(Q(user=user), ~Q(status='disabled'))
        for row in totp_row:
            if row.status == 'enabled' or row.status == 'backup_used':
                return None
            # If status is initialized, then we delete the old totp object
            # This is to prevent multiple totp objects for the same user
            if row.status == 'initialized':
                row.delete()
        # Create a new totp object
        key = pyotp.random_base32()
        backup_codes = generate_backup_codes()
        totp = self.create(
            user=user,
            key=encrypt(key),
            backup_codes=encrypt(backup_codes)
        )
        return key, backup_codes, totp

    def get_totp(self, user):
        try:
            return self.get(Q(user=user), ~Q(status='disabled'))
        except Exception as e:
            return None

    def authenticate(self, user, token, totp_row=None):
        """Authenticate the user using the 6 or 8 length token.

        Exception:
            Throws an exception if the user does not have TOTP enabled/initialized/backup_used.

        Args:
            user (User): Model object
            token (str): 6 len OTP or 8 len backup code
            totp_row (Totp) (optional): Only provide if you are sure that the totp
                             belongs to the same user.
                             Otherwise, it is recommended to not provide this argument.

        Returns:
            Totp or None: Returns Totp when authenticated, None otherwise
        """
        # Check based on the length of the token
        if len(token) == 6:
            # OTP is being used
            if totp_row is None:
                totp_row = self.get_totp(user)  # Get the totp object
                if totp_row is None:
                    return None
            # Get the key from the totp object
            key = decrypt(totp_row.key)
            # Verify the token using the key
            if pyotp.TOTP(key).verify(token):
                # Clear out any backup code attempts
                if totp_row.bc_attempts > 0:
                    totp_row.bc_attempts = 0
                    totp_row.bc_timeout = None
                    totp_row.save()
                # Return the totp object
                # The OTP is authenticated
                return totp_row
        elif len(token) == 8:
            # Backup code is being used
            if totp_row is None:
                totp_row = self.get_totp(user)
                if totp_row is None:
                    return None
            # Check if backup code attempts are timed out
            if totp_row.bc_timeout is not None and now() < totp_row.bc_timeout:
                # Here we assume that the user user is compromised
                # therefore, we simply return None indicating TOTP failure.
                # We also assume that the actual user will use the OTP to login
                # and the possibility for user to use the backup code at the same
                # time is very low.
                return None
            # Only accept backup token if the status is enabled or backup_used
            if totp_row.status == 'enabled' or totp_row.status == 'backup_used':
                # Check if the token is a backup code
                backup_codes = decrypt(totp_row.backup_codes)
                if token in backup_codes:
                    backup_codes.remove(token)
                    totp_row.backup_codes = encrypt(backup_codes)
                    totp_row.status = 'backup_used'
                    totp_row.bc_attempts = 0
                    totp_row.bc_timeout = None
                    totp_row.updated_at = now()
                    totp_row.save()
                    return totp_row
                else:
                    # Increment the backup code attempts
                    # If more than 3, then set the timeout
                    totp_row.bc_attempts += 1
                    if totp_row.bc_attempts > self.BC_ATTEMPTS:
                        totp_row.bc_attempts = 0  # Refresh attempts as timeout is issued
                        totp_row.bc_timeout = now() + self.BC_TIMEOUT
                    totp_row.save()
                    return None
        else:
            return None

    def disable_totp(self, user):
        try:
            self.get(Q(user=user), ~Q(status='disabled')
                     ).update(status='disabled')
        except Exception as e:
            pass

    def create_new_backup_codes(self, user):
        """Create new backup codes for the user. Invalidates old backup codes.

        Args:
            user (User)

        Returns:
            list[str] or None: New backup codes
        """
        try:
            backup_codes = generate_backup_codes()
            # Get the user's totp object
            totp_row = self.get(Q(user=user), ~Q(status='disabled'))
            # Update the backup codes
            totp_row.backup_codes = encrypt(backup_codes)
            totp_row.status = 'enabled'
            totp_row.updated_at = now()
            totp_row.save()
            # Return the new backup codes
            return backup_codes
        except Exception as e:
            return None



##########################################################
#
#                   Outward facing API
#
# This file should ideally be the only export from this
# module/app. (Except permissions.py, middlewares.py)
#
# Using imports from other files may breach security.
#
# If you need to use any other function from this package,
# consider making a new function here and calling that.
#
##########################################################

from .models import AppToken
from .serializers import AppTokenSerializer, UserAppTokenSerializer


def create_app_token(user, request) -> tuple[str, AppToken]:
    """
    Create a new app_token for the user

    Args:
        user: User
        request: HttpRequest

    Returns: tuple[key, app_token]
        key: str
        app_token: AppToken
    """
    return AppToken.objects.create_app_token(user, request)


def delete_app_token(user, app_token_id):
    """
    Delete all app_tokens for the user

    Args:
        user: User
        app_token_id: uuid

    Returns: None
    """
    AppToken.objects.delete_app_token(user, app_token_id)


def get_user(app_token_id):
    """
    Get the user for the app_token

    Args:
        app_token_id: uuid
    """
    return AppToken.objects.get_user(app_token_id)


def get_last_token_session_details(user):
    """
    Get the last app token session details for the user

    Args:
        user: User
    """
    apptoken = AppToken.objects.get_last_session(user)
    if apptoken is not None:
        return UserAppTokenSerializer(apptoken).data
    return apptoken


def get_app_token_if_valid(user, app_token_id, serialized=False):
    """
    Get app_token for the user given the app_token_id

    Args:
        user: User
        app_token_id: uuid
    """
    app_token = AppToken.objects.get_app_token_if_valid(user, app_token_id)
    if app_token is not None and serialized:
        return AppTokenSerializer(app_token).data
    return app_token


def get_all_app_tokens(user, serialized=False):
    app_tokens = AppToken.filter(user=user)
    if serialized:
        return AppTokenSerializer(app_tokens, many=True).data
    return app_tokens


def get_all_active_app_tokens(user, serialized=False):
    app_tokens = AppToken.filter(user=user, is_valid=True)
    if serialized:
        return AppTokenSerializer(app_tokens, many=True).data
    return app_tokens


def delete_all_app_tokens(user):
    AppToken.filter(user=user, is_valid=True).delete()
    return None


def delete_all_app_tokens_except(user, app_token_id):
    AppToken.filter(user=user, is_valid=True).exclude(pk=app_token_id).delete()
    return None



from django.utils.encoding import force_str
from rest_framework import serializers
from .models import User, UserPasswordChange
import re

# Validators


def validate_email(value):
    # Check if email is valid
    regex = r'\b[A-Za-z0-9._%+-]+@[A-Za-z0-9.-]+\.[A-Z|a-z]{2,7}\b'
    if not re.match(regex, value):
        raise serializers.ValidationError('Email is not valid.')
    # Lowercase email
    value = value.lower()
    return value


def validate_password(value):
    # Check if password is at least 8 characters long and contains at least one digit, one uppercase letter,
    # one lowercase letter, and one special character
    regex = r'^(?=.*?[A-Z])(?=.*?[a-z])(?=.*?[0-9])(?=.*?[~#?!@$`\'":;.,%^&*-_+=<>|\/\{\}\[\]\(\)]).{8,}$'
    if not re.match(regex, value):
        raise serializers.ValidationError(
            'Password must be at least 8 characters long and contain at least one digit, one uppercase alphabet, '
            'one lowercase alphabet, and one special character.')
    return value


# Serializers

class UserSerializer(serializers.ModelSerializer):
    class Meta:
        model = User
        fields = ['id', 'is_active', 'created_at', 'first_name', 'last_name', 'timezone',
                  'updated_at', 'email', 'password', 'created_by', 'updated_by']
        extra_kwargs = {'password': {'write_only': True, 'required': True, 'validators': [validate_password]},
                        'created_at': {'read_only': True},
                        'email': {'required': True, 'validators': [validate_email]},
                        'first_name': {'required': True},
                        'id': {'required': False},
                        'last_name': {'required': True}}


class RegistrationSerializer(serializers.Serializer):
    email = serializers.EmailField(validators=[validate_email])
    first_name = serializers.CharField()
    last_name = serializers.CharField()
    password = serializers.CharField(validators=[validate_password])

    def validate(self, data):
        # Check if email, first name, last name, and password are provided
        if 'email' not in data or 'first_name' not in data or 'last_name' not in data or 'password' not in data:
            raise serializers.ValidationError(
                'All fields are required.')
        # Check if email is already in use
        try:
            User.objects.get(email=force_str(data['email']))
            raise serializers.ValidationError('Email is already in use.')
        except User.DoesNotExist:
            pass
        user = User.objects.create_user(
            email=force_str(data['email']),
            first_name=force_str(data['first_name']),
            last_name=force_str(data['last_name']),
            password=force_str(data['password'])
        )
        return user


class LoginSerializer(serializers.Serializer):
    email = serializers.CharField()
    password = serializers.CharField()

    def validate(self, data):
        # Check if email and password are provided
        if 'email' not in data or 'password' not in data:
            raise serializers.ValidationError(
                'All fields are required.')
        try:
            user = User.objects.get(email=force_str(data['email']))
        except User.DoesNotExist:
            user = None
        if user:
            # Check if user is active
            if not user.is_active:
                raise serializers.ValidationError(
                    'Account is disabled. Please contact your administrator.')
            # Check if password is correct
            if not user.check_password(force_str(data['password'])):
                raise serializers.ValidationError('Credentials are invalid.')
        else:
            raise serializers.ValidationError('Credentials are invalid.')
        return user


class PasswordSerializer(serializers.Serializer):
    password = serializers.CharField()

    def validate(self, data):
        # Check if password is provided
        if 'password' not in data:
            raise serializers.ValidationError('Password is required.')
        # validate password
        return validate_password(force_str(data['password']))


class AddUserSerializer(serializers.Serializer):
    email = serializers.EmailField(validators=[validate_email])
    first_name = serializers.CharField()
    last_name = serializers.CharField()
    password = serializers.CharField(validators=[validate_password])

    def validate(self, data):
        # Check if email, first name, last name, and password are provided
        if 'email' not in data or 'first_name' not in data or 'last_name' not in data or 'password' not in data:
            raise serializers.ValidationError(
                'All fields are required.')
        return data


class UserPasswordChangeSerializer(serializers.ModelSerializer):
    class Meta:
        model = UserPasswordChange
        fields = [
            'date_last_changed_by_user',
            'date_last_changed_by_admin',
            'last_changed_by_admin',
            'last_pswd_change_method_by_user',
            'pswd_change_lock_til',
        ]



from ..users_sessions.utils import get_active_session
from ..users_app_tokens.utils import get_active_token


def get_active_user(request):
    """
    Get the active user from the request

    Args:
        request: HttpRequest

    Returns: User
    """
    session = get_active_session(request)
    if session:
        return session.user
    token = get_active_token(request)
    if token:
        return token.user
    return None



from rest_framework import permissions
from .users_sessions.utils import get_active_session
from .users_app_tokens.utils import get_active_token


class IsAuthenticated(permissions.BasePermission):
    message = 'Unauthorized!'
    """
    Permission to check if user session is present
    """

    def has_permission(self, request, view):
        session = get_active_session(request)
        token = get_active_token(request)
        return True if (session and session.user.is_active) or (token and token.user.is_active) else False


class IsAuthenticatedSessionCookie(permissions.BasePermission):
    message = 'Unauthorized!'
    """
    Permission to check if user session is present
    """

    def has_permission(self, request, view):
        session = get_active_session(request)
        return True if session and session.user.is_active else False


class IsAuthenticatedAppToken(permissions.BasePermission):
    message = 'Unauthorized!'
    """
    Permission to check if user session is present
    """

    def has_permission(self, request, view):
        session = get_active_token(request)
        return True if session and session.user.is_active else False



class LogResponse:
    def __init__(self, status, message=None):
        self.status = status
        self.message = message

    def serialize(self):
        result = {"s": self.status}
        if self.message:
            result["m"] = self.message
        return result



import json
from .models import ApiCallLog
from .log_response import LogResponse
from django_axor_auth.configurator import config


class APILogMiddleware:
    def __init__(self, get_response):
        self.get_response = get_response

    def __call__(self, request):
        # Code to be executed for each request before
        # the view (and later middleware) are called.
        response = self.get_response(request)
        # Code to be executed for each request/response after
        # the view is called.

        # Don't run on non-API requests
        if not request.path.startswith(config.URI_PREFIX):
            return response
        # Create log
        try:
            content = json.loads(response.content.decode('utf8'))
        except json.JSONDecodeError:
            # If response is not JSON or empty
            content = dict()
        if response.status_code > 399:
            if 'instance' in content:
                # Omit instance as it is from ErrorMessage class
                # instance also carries the URL that caused the error
                # but this information is already in model's URL column
                content.pop('instance')
            if 'status' in content:
                # Omit status as it is already in model's status column
                content.pop('status')
            log = LogResponse(status=response.status_code, message=content)
        else:
            # check if content has key 'session_id', that means this
            # response is from login or signup API
            if 'user' in content:
                log = LogResponse(status=response.status_code, message=dict(
                    user=content['user']['email']))
            else:
                log = LogResponse(status=response.status_code)
        ApiCallLog.objects.create_log(
            request=request,
            response=log.serialize(),
        )
        return response



import json
from django.db import models
from ..users.users_sessions.utils import get_active_session
from ..users.users_app_tokens.utils import get_active_token


class LogManager(models.Manager):
    def __init__(self):
        super().__init__()

    def create_log(self, request, response):
        """Create a Log entry in the database

        Args:
            request: HTTP request object
            response (dict): Response in format of LogResponse.serialize()
            user (User, optional): User who is performing the action. Defaults to logged-in user or None.
        """
        status = response['s']
        response.pop('s', None)
        session = get_active_session(request)
        app_token = get_active_token(request)
        log = self.model(
            url=request.get_full_path(),
            status=status,
            context=json.dumps(response['m'] if hasattr(
                response, 'm') else response),
            session=session.id if session is not None else None,
            app_token=app_token.id if app_token is not None else None,
            ip=getClientIP(request),
            ua=getUserAgent(request)
        )
        log.save()


# Get Client IP Address
def getClientIP(request):
    x_forwarded_for = request.META.get('HTTP_X_FORWARDED_FOR')
    if x_forwarded_for:
        ip = x_forwarded_for.split(',')[0]
    else:
        ip = request.META.get('REMOTE_ADDR')
    return ip


# Get User Agent
def getUserAgent(request):
    return request.META.get('HTTP_USER_AGENT')



from .models import Session
from rest_framework import serializers


class SessionSerializer(serializers.ModelSerializer):
    class Meta:
        model = Session
        fields = ['id', 'key', 'user', 'expire_at',
                  'is_valid', 'created_at', 'updated_at', 'ip', 'ua']
        extra_kwargs = {
            'key': {'required': True, 'write_only': True},
            'user': {'required': True},
        }


class UserSessionSerializer(serializers.ModelSerializer):
    class Meta:
        model = Session
        fields = ['id', 'is_valid', 'created_at', 'updated_at', 'ip', 'ua']



import json
from Crypto.Cipher import AES
from django.conf import settings


def encrypt(obj):
    cipher = AES.new(key_as_bytes(), AES.MODE_EAX)

    # Encrypt
    ciphertext, tag = cipher.encrypt_and_digest(str.encode(json.dumps(obj)))
    encrypted = [x for x in (cipher.nonce, tag, ciphertext)]
    return b''.join(encrypted)


def decrypt(obj):
    nonce, tag, ciphertext = obj[:16], obj[16:32], obj[32:]
    try:
        cipher = AES.new(key_as_bytes(), AES.MODE_EAX, nonce)
        data = cipher.decrypt_and_verify(ciphertext, tag)
        return json.loads(data)
    except ValueError:
        return None


def key_as_bytes():
    bytes = settings.SECRET_KEY.encode('utf-8')[0:32]
    # If the key is less than 32 bytes, pad it
    while len(bytes) < 32:
        bytes += bytes
    return bytes[0:32]



from django.urls import path
from . import views

"""
/api/user/forgot_password/
"""
urlpatterns = [
    path('request/', views.forgot_password),
    path('validate/', views.check_health),
    path('reset/', views.reset_password),
]



from django.core.mail import send_mail, get_connection
from django_axor_auth.configurator import config


def email_send_helper(email_to: str, subject: str, template: str) -> int:
    """ Sends email to the given email address with the given subject and template

    Args:
        email_to (str): Address to send the email to
        subject (str): Subject of the email
        template (str): HTML template to send (See utils/emailing/base_template.py)

    Returns:
        int: 1 = Success, 0 = Failure
    """
    # Send this token
    try:
        with getEmailConnection() as connection:
            html_message = template
            from_email = config.SMTP_DEFAULT_SEND_FROM
            return send_mail(subject=subject,
                             html_message=html_message,
                             from_email=from_email,
                             recipient_list=[email_to,],
                             connection=connection)
    except Exception as e:
        return 0


def getEmailConnection():
    if config.SMTP_USE_TLS:
        return get_connection(
            host=config.SMTP_HOST,
            port=config.SMTP_PORT,
            username=config.SMTP_USER,
            password=config.SMTP_PASSWORD,
            use_tls=True,
        )
    else:
        return get_connection(
            host=config.SMTP_HOST,
            port=config.SMTP_PORT,
            username=config.SMTP_USER,
            password=config.SMTP_PASSWORD,
            use_ssl=True,
        )



from django.db import models
from django.utils.timezone import now
from .managers import LogManager


class ApiCallLog(models.Model):
    # API URL that the user is trying to access
    url = models.CharField(max_length=255)
    # Response brief for success or error
    # Example: {"status": 200, "message": "Success"}
    # Use LogResponse.serialize() to serialize useful information
    context = models.JSONField(null=True, blank=True)
    # Active session when user is performing the action
    # For login and signup, session_id is None. Instead,
    # 'context' column will have 'user' key.
    session = models.UUIDField(null=True, blank=True)
    app_token = models.UUIDField(null=True, blank=True)
    status = models.IntegerField()
    ip = models.GenericIPAddressField(null=True, blank=True)
    ua = models.TextField(null=True, blank=True)
    created_at = models.DateTimeField(default=now)

    class Meta:
        db_table = 'axor_api_call_logs'
        ordering = ['-created_at']

    objects = LogManager()

    def __str__(self):
        return 'ID ' + str(self.pk)



from django.utils.encoding import force_str
from django.conf import settings
from rest_framework.decorators import api_view, permission_classes
from rest_framework.response import Response
from django_axor_auth.utils.error_handling.error_message import ErrorMessage
from django_axor_auth.middlewares import is_web
from django_axor_auth.configurator import config
# JWT
import jwt
# Session Imports
from .users_sessions.api import create_session, delete_session, get_last_session_details
from .users_sessions.utils import get_active_session
# App Token Imports
from .users_app_tokens.api import create_app_token, get_last_token_session_details, delete_app_token
from .users_app_tokens.utils import get_active_token
# TOTP Imports
from .users_totp.api import has_totp, authenticate_totp
# User Imports
from .serializers import UserSerializer, LoginSerializer, RegistrationSerializer
from .permissions import IsAuthenticated


@api_view(['POST'])
def register(request):
    # Validate request data and create user
    serializer = RegistrationSerializer(data=request.data)
    if serializer.is_valid():
        try:
            # login user
            return login(request._request)
        except Exception as e:
            # Return error message
            err_msg = ErrorMessage(
                detail=str(e),
                status=400,
                instance=request.build_absolute_uri(),
                title='Invalid information provided.',
                code='InvalidRegistrationInfo'
            )
            return err_msg.to_response()
    # Return error message
    errors = serializer.errors
    err_msg = ErrorMessage(
        detail=errors,
        status=400,
        instance=request.build_absolute_uri(),
        title='Invalid information provided.',
        code='InvalidRegistrationInfo'
    )
    return err_msg.to_response()


@api_view(['POST'])
def login(request):
    # Validate request data
    serializer = LoginSerializer(data=request.data)
    if serializer.is_valid():
        # Get user
        user = serializer.validated_data
        # Check if user hash TOTP enabled
        totp_row = has_totp(user)
        if totp_row is not None:
            # If totp code is not provided
            if 'code' not in request.data or ('code' in request.data and (request.data['code'] == None or request.data['code'] == '')):
                return ErrorMessage(
                    detail="TOTP code is required.",
                    status=401,
                    instance=request.build_absolute_uri(),
                    title='2FA code is required',
                    code='TOTPRequired'
                ).to_response()
            # Authenticate TOTP
            if not authenticate_totp(user, force_str(request.data['code']), totp_row):
                return ErrorMessage(
                    detail="Provided TOTP code or backup code is incorrect. Please try again.",
                    status=401,
                    instance=request.build_absolute_uri(),
                    title='2FA code is incorrect',
                    code='TOTPIncorrect'
                ).to_response()
        # Get last session details
        last_session = get_last_session_details(user)  # already serialized
        last_token_session = get_last_token_session_details(
            user)  # already serialized
        # Respond depending on the client
        if is_web(request):
            # Session based authentication
            key, session = create_session(user, request)
            # Add HTTPOnly cookie
            response = Response(data={
                "last_session": last_session,
                "last_token_session": last_token_session,
                "user": UserSerializer(user).data
            },
                status=200
            )
            response.set_cookie(
                key=config.AUTH_COOKIE_NAME,
                value=jwt.encode(
                    {
                        "session_key": key
                    },
                    settings.SECRET_KEY,
                    algorithm='HS256'
                ),
                expires=session.expire_at,
                httponly=True,
                secure=config.AUTH_COOKIE_SECURE,
                samesite=config.AUTH_COOKIE_SAMESITE,
                domain=config.AUTH_COOKIE_DOMAIN
            )
            return response
        else:
            # Token based authentication
            token, app_token = create_app_token(user, request)
            # Respond with token and user data
            return Response(data={
                "last_session": last_session,
                "last_token_session": last_token_session,
                "user": UserSerializer(user).data,
                "session": dict(
                    id=app_token.id,
                    key=jwt.encode(
                        {
                            "app_token": token
                        },
                        settings.SECRET_KEY,
                        algorithm='HS256'
                    ),
                )
            },
                status=200
            )
    errors = serializer.errors
    err_msg = ErrorMessage(
        detail=errors,
        status=400,
        instance=request.build_absolute_uri(),
        title='Invalid credentials',
        code='LoginSerializerErrors'
    )
    return err_msg.to_response()


@api_view(['POST'])
@permission_classes([IsAuthenticated])
def logout(request):
    # Delete session or token
    if is_web(request):
        delete_session(get_active_session(request).user,
                       get_active_session(request).id)
        response = Response(status=200)
        response.delete_cookie(
            key=config.AUTH_COOKIE_NAME,
            domain=config.AUTH_COOKIE_DOMAIN
        )
        return response
    else:
        delete_app_token(get_active_token(request).user,
                         get_active_token(request).id)
    # Return response
    return Response(status=200)


@api_view(['GET'])
@permission_classes([IsAuthenticated])
def me(request):

    # Session-based authentication
    if is_web(request):
        session = get_active_session(request)
        if session is not None:
            return Response(data=dict(
                user=UserSerializer(session.user).data
            ), status=200)
    # Token-based authentication
    app_token = get_active_token(request)
    if app_token is not None:
        return Response(data=dict(
            user=UserSerializer(app_token.user).data
        ), status=200)
    # No valid active session or token found
    return ErrorMessage(
        detail='No active session or token found.',
        status=400,
        instance=request.build_absolute_uri(),
        title='Invalid request',
        code='NoActiveSessionOrToken'
    ).to_response()



from django.contrib import admin
from .models import User

# Register your models here.


class UserAdmin(admin.ModelAdmin):
    list_display = ('email', 'first_name', 'last_name', 'is_active')
    search_fields = ['first_name', 'last_name',
                     'email']
    search_help_text = 'Search with first name, last name, or email address.'
    sortable_by = ['first_name', 'last_name', 'email']
    list_per_page = 50
    ordering = ('-created_at',)
    save_on_top = False
    save_as = False
    readonly_fields = ['created_at', 'password']

    def has_delete_permission(self, request, obj=None):
        return False

    def has_add_permission(self, request):
        return False


admin.site.register(User, UserAdmin)



import json
from django.http import JsonResponse
from django.conf import settings
from ..utils.error_handling.error_message import ErrorMessage
from ..configurator import config


class ValidateJsonMiddleware:
    """Check if the request body data is valid JSON.

    This only works for POST, PUT, and PATCH requests. If the request body
    is not a valid JSON, the request is halted and an error message is
    given as response.
    """

    def __init__(self, get_response):
        self.get_response = get_response

    def __call__(self, request):
        # Run on API requests only
        if request.path.startswith(config.URI_PREFIX) and request.content_type == 'application/json':
            if not self.validate_json_payload(request):
                error = ErrorMessage(
                    detail='Invalid JSON payload.',
                    status=400,
                    instance=request.build_absolute_uri(),
                    title='Invalid data provided'
                )
                return JsonResponse(
                    error.serialize(),
                    status=400)

        response = self.get_response(request)
        return response

    def validate_json_payload(self, request):
        if request.method in ['POST', 'PUT', 'PATCH']:
            try:
                request.data = json.loads(request.body)
                return True
            except json.JSONDecodeError:
                return False
        return True


class VerifyRequestOriginMiddleware:
    """ Verifies if a request origin is allowed to make requests.

    For an origin to make request, it has to be in settings.ALLOW_ORIGINS list.
    """

    def __init__(self, get_response):
        self.get_response = get_response

    def __call__(self, request):
        # Run on API requests only
        if request.path.startswith(config.URI_PREFIX):
            # If Token authentication is used that means
            # the API is accessed by a non-web application.
            # In that case, we don't need to check the origin.
            # OTHERWISE check the origin header!
            if request.active_token is None:
                if not self.validate_origin_header(request):
                    error = ErrorMessage(
                        detail='You are not authorized to perform this action.',
                        status=400,
                        instance=request.build_absolute_uri(),
                        title='Action not allowed.'
                    )
                    return JsonResponse(
                        error.serialize(),
                        status=400)

        response = self.get_response(request)
        return response

    def validate_origin_header(self, request):
        # If DEBUG is True, allow all origins
        if settings.DEBUG:
            return True
        # If ALLOW_ORIGINS is set to '*', allow all origins
        if len(settings.ALLOW_ORIGINS) > 0 and settings.ALLOW_ORIGINS[0] == '*':
            return True
        # Check if the request origin is in ALLOW_ORIGINS
        if request.headers.get('Origin') in settings.ALLOW_ORIGINS:
            return True
        return False



from django.db import models
from django.utils.timezone import now
from django.utils.http import urlsafe_base64_decode, urlsafe_base64_encode
from datetime import timedelta
from .utils import generate_session_key, hash_this, getClientIP, getUserAgent


class SessionManager(models.Manager):
    """Manages the sessions for the users
    - Create a session: Key generation and expiry
    - Delete a session: Invalidating sessions
    - Get a session: Retrieve a session by key, user, or id
    """

    def __init__(self):
        super().__init__()
        # Session expiry in days
        self.KEY_EXPIRE_IN_DAYS = 30
        # Refresh the session before it expires during authentication
        self.KEY_REFERSH_IN_DAYS = 14

    def create_session(self, user, request):
        """Primary use case: Create a session on signup/login

        Args:
            user (User): User to create session for
            request (HttpRequest): For IP and User-Agent

        Returns:
            tuple(key: str, session: Session)
        """
        # Disable previous sessions from same IP and User-Agent
        self.filter(user=user, ip=getClientIP(request),
                    ua=getUserAgent(request), is_valid=True).update(is_valid=False)
        # Create a new session
        key = generate_session_key()
        session = self.create(
            user=user,
            key=hash_this(key),
            ip=getClientIP(request),
            ua=getUserAgent(request),
            expire_at=now() + timedelta(days=self.KEY_EXPIRE_IN_DAYS)
        )
        return urlsafe_base64_encode(key.encode("ascii")), session

    def delete_session(self, user, session_id):
        """Primary use case: Logout a user

        Args:
            user (User): Model object
            session_id (uuid): The row id of the session
        """
        try:
            session = self.get(user=user, id=session_id, is_valid=True)
            if session.is_valid:
                session.is_valid = False
                session.updated_at = now()
                session.save()
        except Exception as e:
            return None
        return None

    def authenticate_session(self, key, ip, ua):
        """This function authenticates a user request

        Args:
            key (str): Session key

        Returns:
            Session or None
        """
        try:
            key = urlsafe_base64_decode(key).decode("ascii")
            session = self.select_related('user').get(
                key=hash_this(key),
                ip=ip,
                ua=ua,
                is_valid=True
            )
            if not session.is_valid:
                return None
            if session.expire_at < now():
                session.is_valid = False
                session.updated_at = now()
                session.save()
                return None
            elif session.expire_at + timedelta(days=self.KEY_REFERSH_IN_DAYS) < now():
                # If the session is about to expire, refresh it
                session.expire_at = now() + timedelta(days=self.KEY_EXPIRE_IN_DAYS)
                session.updated_at = now()
                session.save()
            return session
        except Exception as e:
            return None

    def get_session_if_valid(self, user, session_id):
        """Get session if it is valid and not expired.

        Args:
            user (User): User model object
            session_id (int): Session row id

        Returns:
            Session or None
        """
        try:
            session = self.select_related('user').get(
                user=user,
                id=session_id,
                is_valid=True
            )
            if session.expire_at < now():
                session.is_valid = False
                session.save()
                return None
            return session
        except Exception as e:
            return None

    def get_last_session(self, user):
        return self.filter(user=user).order_by('-created_at').first()

    def get_user(self, session_id):
        try:
            return self.get(id=session_id).user
        except Exception as e:
            return None



from datetime import timedelta
from django.db import models
from django.utils.timezone import now
from .utils import hash_this
from .managers import ForgotPasswordManager
from ..models import User
from django_axor_auth.configurator import config


class ForgotPassword(models.Model):
    key = models.CharField(max_length=150)
    user = models.ForeignKey(User, on_delete=models.CASCADE)
    is_valid = models.BooleanField(default=True)
    is_used = models.BooleanField(default=False)
    ip = models.GenericIPAddressField()
    ua = models.TextField()
    created_at = models.DateTimeField(default=now)
    updated_at = models.DateTimeField(default=now)

    class Meta:
        db_table = 'axor_forgot_password'
        ordering = ['-created_at']

    objects = ForgotPasswordManager()

    def __str__(self):
        return self.user.email

    def set_key(self, key):
        self.key = hash_this(key)
        self.save()

    def check_key(self, key):
        return hash_this(key) == self.key

    def check_valid(self):
        # Check if used
        if self.is_used:
            return False
        # Check if key was created more than x minutes ago
        if (now() - self.created_at) > timedelta(minutes=config.FORGET_PASSWORD_LINK_TIMEOUT):
            self.is_valid = False
            self.save()
            return False
        return self.is_valid

    def set_used(self):
        self.is_used = True
        self.is_valid = False
        self.save()



import jwt
from django.conf import settings
from django.utils.encoding import force_str
from .users_app_tokens.models import AppToken
from .users_sessions.models import Session
from .users_sessions.utils import getClientIP, getUserAgent
from django_axor_auth.configurator import config


class ActiveUserMiddleware:
    """This reads the Cookie or Authorization header and authenticates the user.

    For Cookie-based authentication, `request.active_session` gets be attached to the request.

    For Authorization header-based authentication, `request.active_token` gets be attached to the request.

    To get the authenticated user, use `django-axor-auth.users.api.get_request_user(request)` method.
    """

    def __init__(self, get_response):
        self.get_response = get_response
        self.COOKIE_NAME = config.AUTH_COOKIE_NAME

    def __call__(self, request):
        app_token = self.get_active_app_token(request)
        session = self.get_active_session(request)
        # Attach the session to the request
        request.active_token = app_token
        request.active_session = session

        response = self.get_response(request)
        return response

    def get_active_app_token(self, request):
        apptoken = None
        # Check if auth token is present in header
        if apptoken is None and 'Authorization' in request.headers:
            token = force_str(request.headers['Authorization'])
            if token is None:
                return None
            token = token.split(' ')[-1]  # Get the token from 'Bearer <token>'
            try:
                decoded_jwt_token = jwt.decode(
                    token, settings.SECRET_KEY, algorithms=['HS256'])
                token = decoded_jwt_token.get('app_token')
                return AppToken.objects.authenticate_app_token(token, getUserAgent(request))
            except jwt.InvalidSignatureError:
                apptoken = None
        return apptoken

    def get_active_session(self, request):
        session = None
        # Check if auth token is present in cookies
        try:
            key = request.COOKIES.get(self.COOKIE_NAME)
            try:
                decoded_jwt_key = jwt.decode(
                    key, settings.SECRET_KEY, algorithms=['HS256'])
            except jwt.InvalidSignatureError:
                return None
            key = decoded_jwt_key.get('session_key')
            session = Session.objects.authenticate_session(
                key, getClientIP(request), getUserAgent(request))
            return session
        except (KeyError, Exception) as e:
            session = None
        return session



from datetime import datetime
from django_axor_auth.configurator import config


def base_template(subject, headline, message: list[str], button_text=None, button_link=None):
    """ Base template for email

    Args:
        `subject` (str): Subject of the email. Used in <title> tag.
        `headline` (str): Headline that is displayed in <h1> tag.
        `message` (list[str]): An array of messages that are displayed in <p> tags. The order is important.
        `button_text` (str): Optional. Text for the button
        `button_link` (str): Optional. Link for the button
    """
    company_name = config.APP_NAME
    footer_text = "We may send emails in future if they be beneficial for you or to convey any policy changes. We will never send marketing emails."
    footer_links = [
        {
            "text": "Website",
            "link": config.FRONTEND_URL
        }
    ]
    year = datetime.now().year
    if button_text and button_link:
        btn = f"""
        <br/>
        <table role="presentation" border="0" cellpadding="0" cellspacing="0" class="btn btn-primary">
                <tbody>
                    <tr>
                        <td align="left">
                            <a href="{button_link}" target="_blank">
                                {button_text}
                            </a>
                        </td>
                    </tr>
                </tbody>
            </table>"""
    # Template
    styles = "body,table td{font-size:14px}.body,body{background-color:#f6f6f6}.container,.content{display:block;max-width:580px;padding:10px}.btn a,.btn table td{background-color:#fff}.btn,.btn a,.content,.wrapper{box-sizing:border-box}.align-center,.btn table td,.footer{text-align:center}.clear,.footer{clear:both}img{border:none;-ms-interpolation-mode:bicubic;max-width:100%}body{font-family:Helvetica,sans-serif;-webkit-font-smoothing:antialiased;line-height:1.4;margin:0;padding:0;-ms-text-size-adjust:100%;-webkit-text-size-adjust:100%}table{border-collapse:separate;mso-table-lspace:0pt;mso-table-rspace:0pt;width:100%}table td{font-family:sans-serif;vertical-align:top}.body{width:100%}.container{margin:0 auto!important;width:580px}.btn,.footer,.main{width:100%}.content{margin:0 auto}.main{background:#fff;border-radius:3px}.wrapper{padding:32px}.content-block{padding-bottom:10px;padding-top:10px}.footer{margin-top:10px}.footer a,.footer p,.footer span,.footer td{color:#999;font-size:12px;text-align:center}h1,h2,h3,h4{color:#000;font-weight:400;line-height:1.4;margin:0 0 30px}h1{font-size:35px;font-weight:300;line-height:1.15}ol,p,ul{color:#444;font-size:16px;font-weight:400;margin:0 0 15px}.btn a,a{color:#2372f5;text-decoration:none}ol li,p li,ul li{list-style-position:inside;margin-left:5px}a{font-weight:500}.btn>tbody>tr>td{padding-bottom:15px}.btn table{width:auto}.btn table td{border-radius:5px}.btn a{border:1px solid #2372f5;border-radius:2px;cursor:pointer;display:inline-block;font-size:14px;font-weight:500;margin:0;padding:12px 25px;text-transform:capitalize;letter-spacing:.5px}.btn-primary a{background-color:#2372f5;border-radius:4px;color:#fff;border:none}.last,.mb0{margin-bottom:0}.first,.mt0{margin-top:0}.align-right{text-align:right}.align-left{text-align:left}.text-sm{font-size:13px}.preheader{color:transparent;display:none;height:0;max-height:0;max-width:0;opacity:0;overflow:hidden;mso-hide:all;visibility:hidden;width:0}.powered-by a{text-decoration:none}hr{border:0;border-bottom:1px solid #f6f6f6;margin:20px 0}.logo{font-size:36px;color:#000;letter-spacing:-.75px;font-weight:700;}@media only screen and (max-width:620px){table.body h1{font-size:28px!important;margin-bottom:10px!important}table.body a,table.body ol,table.body p,table.body span,table.body td,table.body ul{font-size:16px!important}table.body .article,table.body .wrapper{padding:10px!important}table.body .content{padding:0!important}table.body .container{padding:0!important;width:100%!important}table.body .main{border-left-width:0!important;border-radius:0!important;border-right-width:0!important}table.body .btn a,table.body .btn table{width:100%!important}table.body .img-responsive{height:auto!important;max-width:100%!important;width:auto!important}}@media all{.ExternalClass{width:100%}.ExternalClass,.ExternalClass div,.ExternalClass font,.ExternalClass p,.ExternalClass span,.ExternalClass td{line-height:100%}.apple-link a{color:inherit!important;font-family:inherit!important;font-size:inherit!important;font-weight:inherit!important;line-height:inherit!important;text-decoration:none!important}#MessageViewBody a{color:inherit;text-decoration:none;font-size:inherit;font-family:inherit;font-weight:inherit;line-height:inherit}.btn-primary a:hover,.btn-primary table td:hover{background-color:#2378fa!important}}"
    template = f"""
    <!DOCTYPE html>
    <html>
    <head>
        <meta name="viewport" content="width=device-width,initial-scale=1">
        <meta http-equiv="Content-Type" content="text/html; charset=UTF-8">
        <title>{subject}</title>
        <style>{styles}</style>
    </head>
    <body>
        <table role="presentation" border="0" cellpadding="0" cellspacing="0" class="body">
            <tr>
                <td>&nbsp;</td>
                <td class="container">
                    <div class="content">
                        <table role="presentation" class="main">
                            <tr>
                                <td class="wrapper">
                                    <table role="presentation" border="0" cellpadding="0" cellspacing="0">
                                        <tr>
                                        <td>
                                            <p class="company-name">
                                                <a href="{config.FRONTEND_URL}" target="_blank" class="logo">{company_name}</a>
                                            </p>
                                            <br/>
                                            <h1>{headline}</h1>
                                            <br/>
                                            {''.join([f'<p>{msg}</p>' for msg in message])}
                                            {btn if button_text and button_link else ''}
                                            <hr/>
                                            <p class="text-sm">{footer_text}</p>
                                            <br/>
                                            <div class="align-center">{company_name} {year}.</div>
                                            <div class="align-center">
                                            {" • ".join([f'<a href="{link["link"]}" target="_blank">{link["text"]}</a>' for link in footer_links])}
                                            </div>
                                        </td>
                                        </tr>
                                    </table>
                                </td>
                            </tr>
                        </table>
                    </div>
                </td>
                <td>&nbsp;</td>
            </tr>
        </table>
    </body>
    </html>
    """
    return template



import uuid
from django.db import models
from django.utils.timezone import now
from .managers import UserManager, UserPasswordChangeManager
import bcrypt


class User(models.Model):
    id = models.UUIDField(unique=True, primary_key=True, default=uuid.uuid4)
    password = models.CharField(max_length=150)
    email = models.CharField(
        max_length=150, unique=True, null=True, blank=True)
    first_name = models.CharField(max_length=150)
    last_name = models.CharField(max_length=150)
    timezone = models.CharField(max_length=150, default='America/Vancouver')
    is_active = models.BooleanField(default=True)
    created_at = models.DateTimeField(default=now)
    created_by = models.ForeignKey(
        'self', on_delete=models.SET_NULL, related_name='user_created_by', null=True, blank=True)
    updated_at = models.DateTimeField(default=now)
    updated_by = models.ForeignKey(
        'self', on_delete=models.SET_NULL, related_name='user_updated_by', null=True, blank=True)

    class Meta:
        db_table = 'axor_users'
        ordering = ['first_name', 'last_name']

    objects = UserManager()

    def __str__(self):
        return self.first_name + ' ' + self.last_name + ' - ' + self.email

    def set_password(self, password):
        self.password = bcrypt.hashpw(password.encode(
            'utf-8'), bcrypt.gensalt()).decode('utf-8')
        self.save()

    def check_password(self, password):
        try:
            return bcrypt.checkpw(password.encode('utf-8'), self.password.encode('utf-8'))
        except Exception as e:
            return False


class UserPasswordChange(models.Model):
    method_choices = (
        ('authenticated', 'authenticated'),
        ('forgot_password', 'forgot_password'),
    )

    user = models.ForeignKey(User, on_delete=models.CASCADE,
                             related_name='pswd_user', null=True, blank=True)
    date = models.DateTimeField(default=None, null=True)
    method = models.CharField(
        max_length=32, choices=method_choices, default=None, null=True)

    class Meta:
        db_table = 'axor_user_password_change'
        ordering = ['date']

    objects = UserPasswordChangeManager()



import string
import random
import hashlib
from secrets import token_urlsafe
from django_axor_auth.configurator import config


def generate_backup_codes():
    """Generate backup codes for the user

    Returns:
        list: A list of backup codes
    """
    backup_codes = []
    for _ in range(config.TOTP_NUM_OF_BACKUP_CODES):
        backup_code = ''.join(random.choices(
            string.ascii_lowercase + string.digits, k=config.TOTP_BACKUP_CODE_LENGTH))
        backup_codes.append(backup_code)
    return backup_codes


def generate_token():
    # This token length can be of any size as sha256 is used
    # and saved in database as fixed 64 characters. At least
    # 64 characters key is recommended.
    return token_urlsafe(64)


# Hash Function for key
def hash_this(string):
    return hashlib.sha256(str(string).encode('utf-8')).hexdigest()


# Get Client IP Address
def getClientIP(request):
    x_forwarded_for = request.META.get('HTTP_X_FORWARDED_FOR')
    if x_forwarded_for:
        ip = x_forwarded_for.split(',')[0]
    else:
        ip = request.META.get('REMOTE_ADDR')
    return ip


# Get User Agent
def getUserAgent(request):
    return request.META.get('HTTP_USER_AGENT')



from django.db import models
from django.utils.timezone import now


class UserManager(models.Manager):
    def __init__(self):
        super().__init__()

    def create_user(self, password, **extra_fields):
        user = self.model(**extra_fields)
        user.set_password(password)
        user.save()
        return user


class UserPasswordChangeManager(models.Manager):
    def __init__(self):
        super().__init__()

    def save_due_forgot_password_form(self, user):
        """Create a row in the UserPasswordChange table for the user who forgot their password.

        Args:
            user (User): User object
        """
        self.create(
            user=user,
            date=now(),
            method='forgot_password',
        )



from django.contrib import admin
from .models import AppToken

# Register your models here.


class AppTokenAdmin(admin.ModelAdmin):
    list_display = ('id', 'user', 'is_valid', 'ip', 'created_at')
    search_fields = ['user__email', 'user__first_name',
                     'user__last_name', 'id', 'ip']
    search_help_text = 'Search with user\'s first and last name, email, session Id, or IP address.'
    list_per_page = 100
    ordering = ('-created_at',)
    save_on_top = False
    save_as = False
    readonly_fields = ['id', 'user', 'token', 'ip', 'ua', 'created_at']

    def has_delete_permission(self, request, obj=None):
        return False

    def has_add_permission(self, request):
        return False


admin.site.register(AppToken, AppTokenAdmin)



import hashlib
from secrets import token_urlsafe

REQUEST_SESSION_FIELD = 'active_session'


def get_active_session(request):
    if hasattr(request, REQUEST_SESSION_FIELD):
        return getattr(request, REQUEST_SESSION_FIELD)
    return None

# Generate session key


def generate_session_key():
    # This token length can be of any size as sha256 is used
    # and saved in database as fixed 64 characters. At least
    # 64 characters key is recommended.
    return token_urlsafe(64)


# Hash Function for key
def hash_this(string):
    return hashlib.sha256(str(string).encode('utf-8')).hexdigest()


# Get Client IP Address
def getClientIP(request):
    x_forwarded_for = request.META.get('HTTP_X_FORWARDED_FOR')
    if x_forwarded_for:
        ip = x_forwarded_for.split(',')[0]
    else:
        ip = request.META.get('REMOTE_ADDR')
    return ip


# Get User Agent
def getUserAgent(request):
    return request.META.get('HTTP_USER_AGENT')



from django.db import models
from ..models import User
from django.utils.timezone import now
from django.db import models
from .managers import TotpManager


class Totp(models.Model):
    STATUS_CHOICES = (
        ('initialized', 'Initialized'),
        ('enabled', 'Enabled'),
        ('disabled', 'Disabled'),
        ('backup_used', 'Backup Used')
    )
    user = models.ForeignKey(User, on_delete=models.CASCADE)
    key = models.BinaryField()
    status = models.CharField(
        max_length=16, choices=STATUS_CHOICES, default='initialized')
    backup_codes = models.BinaryField()
    created_at = models.DateTimeField(default=now)
    updated_at = models.DateTimeField(default=now)
    bc_attempts = models.IntegerField(default=0)
    bc_timeout = models.DateTimeField(null=True, blank=True)

    class Meta:
        db_table = 'axor_totp'
        ordering = ['-created_at']

    objects = TotpManager()

    def __str__(self):
        return f"{self.user}"



from django.db import models
from .utils import generateKey, getClientIP, getUserAgent, hash_this
from ..api import get_user


class ForgotPasswordManager(models.Manager):
    def __init__(self):
        super().__init__()

    def create_forgot_password(self, request, email):
        user = get_user(email)
        if not user:
            return None, None
        # create forgot password instance
        key = generateKey()
        fp = self.create(
            user=user,
            key=hash_this(key),
            ip=getClientIP(request),
            ua=getUserAgent(request)
        )
        return key, fp



from rest_framework.views import exception_handler
from .error_message import ErrorMessage
import json


def drf_exception_handler(exc, context):
    # Call REST framework's default exception handler first,
    # to get the standard error response.
    response = exception_handler(exc, context)

    # Now add the HTTP status code to the response.
    if response is not None:
        error = ErrorMessage(
            title=response.data['detail'],
            detail=response.data['detail'],
            status=response.status_code,
            instance=context['request'].build_absolute_uri(),
            code='DRF001'
        )
        response.data = json.dumps(error.serialize())
        response.content = json.dumps(error.serialize())
        response['Content-Type'] = 'application/problem+json'

    return response



from django.db import models
from django.utils.timezone import now
from .utils import generate_token, hash_this, getClientIP, getUserAgent
from django.utils.http import urlsafe_base64_decode, urlsafe_base64_encode


class AppTokenManager(models.Manager):

    def __init__(self):
        super().__init__()

    def create_app_token(self, user, request):
        """Primary use case: Create a app token on login for Authorization header

        Args:
            user (User): User to create app_token for
            request (HttpRequest): For IP and User-Agent

        Returns:
            tuple(key: str, token: Token)
        """
        # Disable previous sessions from same IP and User-Agent
        self.filter(user=user, ip=getClientIP(request),
                    ua=getUserAgent(request), is_valid=True).update(is_valid=False)
        key = generate_token()
        token = self.create(
            user=user,
            token=hash_this(key),
            ip=getClientIP(request),
            ua=getUserAgent(request),
        )
        return urlsafe_base64_encode(key.encode("ascii")), token

    def delete_app_token(self, user, app_token_id):
        """Primary use case: Logout a user

        Args:
            user (User): Model object
            app_token_id (uuid): The row id of the app_token
        """
        try:
            app_token = self.get(user=user, id=app_token_id, is_valid=True)
            if app_token.is_valid:
                app_token.is_valid = False
                app_token.updated_at = now()
                app_token.save()
        except Exception as e:
            return None
        return None

    def authenticate_app_token(self, token, ua):
        """This function authenticates a user request

        Args:
            token (str): App token from Authorization header

        Returns:
            AppToken or None
        """
        try:
            key = urlsafe_base64_decode(token).decode("ascii")
            app_token = self.select_related('user').get(
                token=hash_this(key),
                ua=ua,
                is_valid=True
            )
            if not app_token.is_valid:
                return None
            return app_token
        except Exception as e:
            return None

    def get_app_token_if_valid(self, user, app_token_id):
        """Get app_token if it is valid and not expired.

        Args:
            user (User): User model object
            app_token_id (int): Session row id

        Returns:
            Session or None
        """
        try:
            app_token = self.select_related('user').get(
                user=user,
                id=app_token_id,
                is_valid=True
            )
            return app_token
        except Exception as e:
            return None

    def get_last_session(self, user):
        return self.filter(user=user).order_by('-created_at').first()

    def get_user(self, app_token_id):
        try:
            return self.get(id=app_token_id).user
        except Exception as e:
            return None



import json
from django.contrib import admin
from .models import ApiCallLog
from ..users.users_app_tokens.api import get_user as get_user_from_app_token
from ..users.users_sessions.api import get_user as get_user_from_session

# Register your models here.


class ApiCallLogAdmin(admin.ModelAdmin):
    list_display = ('id', 'user_email', 'url', 'status', 'ip', 'created_at')
    search_fields = ['id', 'session', 'app_token', 'url', 'ip']
    search_help_text = 'Search with Id, URL and session id, token id, or IP address.'
    list_per_page = 50
    ordering = ('-created_at',)
    save_on_top = False
    save_as = False
    readonly_fields = ['url', 'context', 'session', 'ip', 'ua',
                       'app_token', 'status', 'created_at']

    def user_email(self, obj):
        if obj.app_token:
            user = get_user_from_app_token(obj.app_token)
            if user:
                return user.email
        elif obj.session:
            user = get_user_from_session(obj.session)
            if user:
                return user.email
        else:
            context = json.loads(obj.context)
            if 'm' in context and 'user' in context['m']:
                """This only works for login and sign up requests as the APILogMiddleware
                the user object and store it in log context"""
                return context['m']['user']
        return '---'

    def has_delete_permission(self, request, obj=None):
        return False

    def has_add_permission(self, request):
        return False


admin.site.register(ApiCallLog, ApiCallLogAdmin)



from django.urls import path
from . import views

"""
api/user/totp/
"""
urlpatterns = [
    path('init/', views.totp_init),
    path('enable/', views.totp_enable),
    path('disable/', views.totp_disable),
    path('new-backup-codes/', views.totp_new_backup_codes),
]



from django.utils.encoding import force_str
from django.utils.timezone import now
from rest_framework.decorators import api_view, permission_classes
from rest_framework.response import Response
from django_axor_auth.users.permissions import IsAuthenticated
from django_axor_auth.users.users_utils.active_user import get_active_user
from django_axor_auth.utils.error_handling.error_message import ErrorMessage
from django_axor_auth.configurator import config
from .models import Totp


@api_view(['POST'])
@permission_classes([IsAuthenticated])
def totp_init(request):
    # Get user
    user = get_active_user(request)
    # Create TOTP
    new_totp = Totp.objects.create_totp(user)
    if new_totp is None:
        return ErrorMessage(
            title='TOTP is already enabled',
            detail='You are attempting to setup TOTP but it is already enabled.',
            status=400,
            code='TOTPAlreadyEnabledOnInit',
            instance=request.build_absolute_uri()
        ).to_response()
    key, backup_codes, _ = new_totp
    return Response(data={
        'key': key,
        'backup_codes': backup_codes,
        'provision': f"otpauth://totp/{config.APP_NAME}:{user.email.replace('@','%40')}?secret={key}&issuer={config.APP_NAME}"
    }, status=201)


@api_view(['POST'])
@permission_classes([IsAuthenticated])
def totp_enable(request):
    """ Enable TOTP for the user after user initiates the TOTP.

    Request Parameters:
        code (str): 6 len OTP code only
    """
    token = force_str(request.data.get('code'))
    # If backup code is being used, then return 400 right away
    if len(token) > 6:
        return ErrorMessage(
            title='Invalid TOTP Token',
            detail='Please use a 6-digit TOTP token to enable TOTP. Backup codes cannot be used.',
            status=400,
            code='InvalidTOTPTokenTooLong',
            instance=request.build_absolute_uri()
        ).to_response()
    # Try to authenticate the user
    try:
        # Get user
        user = get_active_user(request)
        # If authenticated, totp object will be returned otherwise None
        totp = Totp.objects.authenticate(user, token)
        # Only enable if the status is initialized
        # Disabled totp's cannot be enabled again
        if totp:
            if totp.status == 'initialized':
                totp.status = 'enabled'
                totp.save()
            return Response(status=200)
        return ErrorMessage(
            title='Invalid TOTP Token',
            detail='The provided TOTP token is invalid. Please try again.',
            status=400,
            code='InvalidTOTPToken',
            instance=request.build_absolute_uri()
        ).to_response()
    except Totp.DoesNotExist:
        return ErrorMessage(
            title='TOTP not set',
            detail='You will need to start the TOTP setup process first.',
            status=400,
            code='TOTPNotSetup',
            instance=request.build_absolute_uri()
        ).to_response()


@api_view(['PUT'])
@permission_classes([IsAuthenticated])
def totp_disable(request):
    """ Disable TOTP for the user.

    Request Parameters:
        code (str): 6 len OTP code or 8 len backup code
    """
    # Get the token from request
    token = force_str(request.data.get('code'))
    # Try to authenticate the user
    totp = Totp.objects.authenticate(get_active_user(request), token)
    if totp:
        totp.status = 'disabled'
        totp.updated_at = now()
        totp.save()
        return Response(status=200)
    return ErrorMessage(
        title='Invalid TOTP Token',
        detail='The provided TOTP token or backup code is invalid. Please try again.',
        status=400,
        code='InvalidTOTPToken',
        instance=request.build_absolute_uri()
    ).to_response()


@api_view(['POST'])
@permission_classes([IsAuthenticated])
def totp_new_backup_codes(request):
    """
    Generate new backup codes for the user. Invalidates the old backup codes.
    """
    # Get the token from request
    token = force_str(request.data.get('code'))
    # Try to authenticate the user
    totp = Totp.objects.authenticate(get_active_user(request), token)
    if totp:
        # Create new backup codes
        backup_codes = Totp.objects.create_new_backup_codes(
            get_active_user(request))
        if backup_codes is None:
            return ErrorMessage(
                title='TOTP not set',
                detail='You will need to start the TOTP setup process first.',
                status=400,
                code='TOTPNotSetup',
                instance=request.build_absolute_uri()
            ).to_response()
        return Response(data={'backup_codes': backup_codes}, status=200)
    return ErrorMessage(
        title='Invalid TOTP Token',
        detail='The provided TOTP token or backup code is invalid. Please try again.',
        status=400,
        code='InvalidTOTPToken',
        instance=request.build_absolute_uri()
    ).to_response()



from .....utils.emailing.base_template import base_template


def new_user_account_template(first_name, url, email_subject='Welcome to Pluto Health'):
    return base_template(
        subject=email_subject,
        headline="Let's get you upto speed!",
        message=[
            'Hi ' + first_name + ',',
            'We are excited to have you onboard.',
            'Use the button below to activate your account.',
        ],
        button_text='Activate Account',
        button_link=url,
    )



from .....utils.emailing.base_template import base_template


def forgot_password_template(first_name, url, ip, email_subject='Reset your password'):
    return base_template(
        subject=email_subject,
        headline="Forgot your password?",
        message=[
            'Hi ' + first_name + ',',
            'We received a request to reset your password. Use the button below to reset your password.',
            '',
            'If you did not request a password reset, please ignore this email.',
            'IP requesting reset: ' + ip,
        ],
        button_text='Reset Password',
        button_link=url,
    )



from django.urls import path, include
from . import views


"""
api/user/
"""
urlpatterns = [
    path('register/', views.register),
    path('login/', views.login),
    path('logout/', views.logout),
    path('me/', views.me),
    path('totp/', include('django_axor_auth.users.users_totp.urls')),
    path('forgot_password/',
         include('django_axor_auth.users.users_forgot_password.urls')),
]



##########################################################
#
#                   Outward facing API
#
# This file should ideally be the only export from this
# module/app. (Except permissions.py, middlewares.py)
#
# Using imports from other files may breach security.
#
# For API requests, use the urls.py file endpoints.
#
# If you need to use any other function from this package,
# consider making a new function here and calling that.
#
##########################################################

from .models import Totp
from django.db.models import Q


def has_totp(user):
    """
    Check if the user has TOTP enabled

    Args:
        user (User): Model object

    Returns: TOTP or None
    """
    try:
        totp = Totp.objects.get(Q(user=user), ~Q(
            status='disabled'), ~Q(status='initialized'))
        return totp
    except Totp.DoesNotExist:
        return None


def authenticate_totp(user, token, totp_row=None):
    """
    Authenticate the user using the 6 or 8 length token.
    Only send request to this method if `has_totp` returns Totp object,
    otherwise it will produce an Exception.

    Args:
        user (User): Model object
        token (str): 6 or 8 length token

    Returns: bool
    """
    if totp_row is None:
        totp_row = has_totp(user)
        if totp_row is None:
            return False
    return True if Totp.objects.authenticate(user, token, totp_row) else False


def disable_totp(user):
    """
    Disable the TOTP for the user

    Args:
        user (User): Model object

    Returns: bool
    """
    return Totp.objects.disable_totp(user)



##########################################################
#
#                   Outward facing API
#
# This file should ideally be the only export from this
# module/app. (Except permissions.py, middlewares.py)
#
# Using imports from other files may breach security.
#
# If you need to use any other function from this package,
# consider making a new function here and calling that.
#
##########################################################

from .models import Session
from .serializers import SessionSerializer, UserSessionSerializer


def create_session(user, request) -> tuple[str, Session]:
    """
    Create a new session for the user

    Args:
        user: User
        request: HttpRequest

    Returns: tuple[key, session]
        key: str
        session: Session
    """
    return Session.objects.create_session(user, request)


def delete_session(user, session_id):
    """
    Delete all sessions for the user

    Args:
        user: User
        session_id: int

    Returns: None
    """
    Session.objects.delete_session(user, session_id)


def get_user(session_id):
    """
    Get the user for the session

    Args:
        session_id: int
    """
    return Session.objects.get_user(session_id)


def get_last_session_details(user):
    """
    Get the last session details for the user

    Args:
        user: User
    """
    session = Session.objects.get_last_session(user)
    if session is not None:
        return UserSessionSerializer(session).data
    return session


def get_session_if_valid(user, session_id, serialized=False):
    """
    Get session for the user given the session_id

    Args:
        user: User
        session_id: int
    """
    session = Session.objects.get_session_if_valid(user, session_id)
    if session is not None and serialized:
        return SessionSerializer(session).data
    return session


def get_all_sessions(user, serialized=False):
    sessions = Session.filter(user=user)
    if serialized:
        return SessionSerializer(sessions, many=True).data
    return sessions


def get_all_active_sessions(user, serialized=False):
    sessions = Session.filter(user=user, is_valid=True)
    if serialized:
        return SessionSerializer(sessions, many=True).data
    return sessions


def delete_all_sessions(user):
    Session.filter(user=user, is_valid=True).update(is_valid=False)
    return None


def delete_all_sessions_except(user, session_id):
    Session.filter(user=user, is_valid=True).exclude(
        pk=session_id).update(is_valid=False)
    return None



from django.contrib import admin
from .models import Totp

# Register your models here.


class TotpAdmin(admin.ModelAdmin):
    list_display = ('user', 'active', 'status', 'created_at')
    search_fields = ['user__email']
    search_help_text = 'Search with user email.'
    list_per_page = 100
    ordering = ('-created_at',)
    save_on_top = False
    save_as = False
    readonly_fields = ['user', 'key', 'backup_codes', 'bc_attempts', 'bc_timeout',
                       'created_at', 'updated_at']

    def active(self, obj):
        return obj.status == 'enabled'

    def has_delete_permission(self, request, obj=None):
        return False

    def has_add_permission(self, request):
        return False


admin.site.register(Totp, TotpAdmin)



from .models import User
from django_axor_auth.middlewares import is_web
# Session Imports
from .users_sessions.utils import get_active_session
# App Token Imports
from .users_app_tokens.utils import get_active_token
from .serializers import UserSerializer


def get_user(email):
    """
    Get active User object

    Args:
        email (str): User email

    Returns: User or None
    """
    try:
        account = User.objects.get(email=email, is_active=True)
        return account
    except User.DoesNotExist:
        return None


def get_request_user(request):
    if is_web(request):
        # Check if session is active
        session = get_active_session(request)
        if session is not None:
            return session.user
    # Check if token is active
    app_token = get_active_token(request)
    if app_token is not None:
        return app_token.user
    return None


def add_user(email, password, first_name, last_name, created_by=None, serialized=False):
    """
    Add a new User

    Args:
        email (str): User email
        password (str): User password
        first_name (str): User first name
        last_name (str): User last name
        created_by (User, optional): User object. Defaults to None.
        serialized (bool, optional): Return serialized User. Defaults to False.

    Returns: User or None
    """
    try:
        account = User.objects.create_user(
            email=email,
            password=password,
            first_name=first_name,
            last_name=last_name,
            created_by=created_by
        )
        if serialized:
            return UserSerializer(data=account).data
        else:
            return account
    except Exception as e:
        raise Exception(e)


def change_password(user, new_password):
    """
    Change user password

    Args:
        user (User): User object
        new_password (str): New password

    Returns: User or None
    """
    try:
        user.set_password(new_password)
        user.save()
        return user
    except Exception as e:
        raise Exception(e)


def change_email(user, new_email):
    """
    Change user email

    Args:
        user (User): User object
        new_email (str): New email

    Returns: User or None
    """
    try:
        user.email = new_email
        user.save()
        return user
    except Exception as e:
        raise Exception(e)


def change_name(user, new_first_name, new_last_name):
    """
    Change user name

    Args:
        user (User): User object
        new_first_name (str): New first name
        new_last_name (str): New last name

    Returns: User or None
    """
    try:
        user.first_name = new_first_name
        user.last_name = new_last_name
        user.save()
        return user
    except Exception as e:
        raise Exception(e)


def disable_user(user: User):
    """
    Disable user

    Args:
        user (User): User object

    Returns: User or None
    """
    try:
        user.is_active = False
        user.save()
        return user
    except Exception as e:
        raise Exception(e)


def enable_user(user):
    """
    Enable user

    Args:
        user (User): User object

    Returns: User or None
    """
    try:
        user.is_active = True
        user.save()
        return user
    except Exception as e:
        raise Exception(e)


def delete_user(user):
    """
    Delete user

    Args:
        user (User): User object

    Returns: User or None
    """
    try:
        user.delete()
        return user
    except Exception as e:
        raise Exception(e)



from .configurator import config


class HeaderRequestedByMiddleware:
    """Check for HTTP_X_REQUESTED_BY header and set request.requested_by accordingly.
    """

    def __init__(self, get_response):
        self.get_response = get_response

    def __call__(self, request):
        # Run on API requests only
        if request.path.startswith(config.URI_PREFIX):
            # If 'X-Requested-By' header is present and not set to 'web'
            # then token based authentication is used. In such case,
            # provide 'Authorization' header with token.
            if 'HTTP_X_REQUESTED_BY' in request.META:
                request.requested_by = request.META['HTTP_X_REQUESTED_BY']
            else:
                request.requested_by = 'web'

        response = self.get_response(request)
        return response


def is_web(request):
    return request.requested_by == 'web'
